import _extends from 'babel-runtime/helpers/extends';
import _Object$keys from 'babel-runtime/core-js/object/keys';
import _typeof from 'babel-runtime/helpers/typeof';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import _Symbol from 'babel-runtime/core-js/symbol';
import invariant from 'invariant';
import { Observable } from '@reactivex/rxjs';

import { Collection } from './Collection';
import { Store } from './Store';
import combineStores from './util/combineStores';
import createGetter from './util/createGetter';
import capitalizeString from './util/capitalizeString';

import isDispatcher from 'fluorine-lib/lib/util/isDispatcher';

import { Iterable, Set } from 'immutable';

var resultCache = _Symbol('resultCache');

export var Orchestra = function () {
  Orchestra.isOrchestra = function isOrchestra(obj) {
    return (typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) === 'object' && obj instanceof Orchestra;
  };

  function Orchestra() {
    _classCallCheck(this, Orchestra);

    for (var _len = arguments.length, stores = Array(_len), _key = 0; _key < _len; _key++) {
      stores[_key] = arguments[_key];
    }

    invariant(stores.length > 0 && stores.every(Store.isStore), 'Orchestra: constructor expects to receive Stores.');

    var _stores = stores.reduce(function (acc, store) {
      var identifier = store.identifier;


      if (acc[identifier]) {
        throw new Error('Orchestra: The identifier `' + identifier + '` is not unique.');
      }

      acc[identifier] = store;
      return acc;
    }, {});

    this[resultCache] = {};
    this.stores = _stores;
    this.externals = {};
  }

  Orchestra.prototype.addReducer = function addReducer(identifier, reducer) {
    invariant(typeof identifier === 'string', 'Orchestra: `identifier` is expected to be a string.');
    invariant(typeof reducer === 'function', 'Orchestra: `reducer` is expected to be a reducer function.');

    var externals = this.externals;
    var stores = this.stores;


    invariant(!stores.hasOwnProperty(identifier), 'Orchesta: The identifier `' + identifier + '` is already taken by a Store.');
    invariant(!externals.hasOwnProperty(identifier), 'Orchesta: The identifier `' + identifier + '` is not unique.');

    externals[identifier] = reducer;
    return this;
  };

  Orchestra.prototype.reduce = function reduce(dispatcher) {
    invariant(isDispatcher(dispatcher), 'Orchestra: `dispatcher` is expected to be a Fluorine dispatcher.');
    if (this[resultCache][dispatcher.identifier]) {
      return this[resultCache][dispatcher.identifier];
    }

    var stores = this.stores;
    var externals = this.externals;

    // Assemble reduced externals

    var _externals = _Object$keys(externals).reduce(function (acc, key) {
      var external = externals[key];

      acc[key] = dispatcher.reduce(external);

      return acc;
    }, {});

    var _stores = _Object$keys(stores).reduce(function (acc, key) {
      var store = stores[key];
      var reducer = store.getReducer();

      acc[key] = dispatcher.reduce(reducer);

      return acc;
    }, {});

    // Empty collections for convenience
    var _emptyCollections = _Object$keys(stores).reduce(function (acc, key) {
      var store = stores[key];
      var collection = store.createCollection();

      acc[key] = collection;
      return acc;
    }, {});

    var _loop = function _loop(identifier) {
      if (stores.hasOwnProperty(identifier)) {
        var _ret2 = function () {
          var _extends2;

          // Get the unresolved fluorine store
          var fluorineStore = _stores[identifier];

          // Get the post hook and all dependencies for this store
          var orchestraStore = stores[identifier];
          var post = orchestraStore.getPost();
          var dependencies = orchestraStore.getDependencies();
          var reducer = orchestraStore.getReducer();

          // If the store doesn't have any dependencies we only apply the post hook
          if (_Object$keys(dependencies).length === 0) {
            // Override the old fluorine store
            _stores[identifier] = fluorineStore.map(function (state) {
              return state.map(post);
            }).publishReplay(1).refCount();

            return {
              v: 'continue'
            }; // Honey, I'm home (early)
          }

          // Build object containing fluorine stores for all dependencies
          var _dependencyStores = _Object$keys(dependencies).reduce(function (acc, dependency) {
            if (_externals[dependency]) {
              acc[dependency] = _externals[dependency];
            } else if (_stores[dependency]) {
              acc[dependency] = _stores[dependency];
            } else {
              throw new Error('Orchestra: Failed to resolve dependency for identifier `' + dependency + '`.');
            }

            return acc;
          }, {});

          // Override the old fluorine store
          _stores[identifier] = combineStores(_extends({}, _dependencyStores, (_extends2 = {}, _extends2[identifier] = fluorineStore, _extends2))).map(function (deps) {
            return _Object$keys(dependencies).reduce(function (state, dependencyIdentifier) {
              // Retrieve the dependency getter, setter and its state
              var dependencyState = deps[dependencyIdentifier];
              var _dependencies$depende = dependencies[dependencyIdentifier];
              var getter = _dependencies$depende.getter;
              var setter = _dependencies$depende.setter;

              var emptyCollection = _emptyCollections[dependencyIdentifier];
              var dependencyName = capitalizeString(dependencyIdentifier);

              var missingIds = [];

              // Modify every item
              var _state = state.map(function (x) {
                // Get dependency ids
                var ids = getter(x, dependencyState);

                invariant(typeof ids === 'string' || Iterable.isIterable(ids) || Array.isArray(ids), 'Orchestra: `ids` is expected to be a string, an iterable, or an Array.');

                // Collect missing ids from dependency item
                if (typeof ids === 'string') {
                  missingIds.push(ids);
                } else {
                  ids.forEach(function (id) {
                    if (!dependencyState.get(id)) {
                      missingIds.push(id);
                    }
                  });
                }

                var _get = createGetter(dependencyState, emptyCollection, ids);
                if (setter) {
                  return setter(x, _get);
                }

                x['get' + dependencyName] = _get;
                return x;
              });

              // Report missing ids
              var dependencyStore = stores[dependencyIdentifier];
              if (dependencyStore) {
                dependencyStore._missing(new Set(missingIds), identifier);
              }

              return _state;
            }, deps[identifier]).map(post);
          }).publishReplay(1).refCount();
        }();

        if ((typeof _ret2 === 'undefined' ? 'undefined' : _typeof(_ret2)) === "object") return _ret2.v;
      }
    };

    for (var identifier in stores) {
      var _ret = _loop(identifier);

      if (_ret === 'continue') continue;
    }

    this[resultCache][dispatcher.identifier] = _stores;
    return _stores;
  };

  return Orchestra;
}();

export default function createOrchestra() {
  for (var _len2 = arguments.length, stores = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    stores[_key2] = arguments[_key2];
  }

  return new (Function.prototype.bind.apply(Orchestra, [null].concat(stores)))();
}